#
# Copyright 2001 by Object Craft P/L, Melbourne, Australia.
#
# LICENCE - see LICENCE file distributed with this software for details.
#
# Standard albatross tags
#
import time

from albatross.template import Content, EmptyTag, EnclosingTag
from albatross.common import *

# Empty content object which tags can use as default value for
# optional child tag content.  <al-head>, <al-tail>, etc.
not_specified = Content()


# Escape text for attribute values
def escape(text):
    text = str(text)
    text = text.replace('&', '&amp;')
    text = text.replace('<', '&lt;')
    text = text.replace('>', '&gt;',)
    text = text.replace('"', '&quot;')
    text = text.replace("'", '&#39;')
    return text


# Implement the expr, nameattr, etc attribute behaviour
class EvalAttribMixin:

    def assert_has_attribs(self, *attrs):
        """ Check that the tag has at least one of the attributes """
        for attr in attrs:
            if self.has_attrib(attr):
                return
        if len(attrs) == 1:
            self.raise_error('missing %s attribute' % repr(attrs[0]))
        else:
            attr_list = ', '.join([repr(attr) for attr in attrs[:-1]]) 
            self.raise_error('missing %s or %s attribute' %\
                             (attr_list, repr(attrs[-1])))

    def eval_attrib(self, ctx, attr, kind = 'eval'):
        code_attr = 'co_' + attr
        try:
            code_obj = getattr(self, code_attr)
        except AttributeError:
            code_obj = compile(self.get_attrib(attr), '<albatross>', kind)
            setattr(self, code_attr, code_obj)
        return ctx.eval_expr(code_obj)
            
    def get_name(self, ctx):
        alias = self.get_attrib('alias')
        if alias:
            name = ctx.make_alias(alias)
        elif self.has_attrib('nameexpr'):
            name = self.eval_attrib(ctx, 'nameexpr')
        else:
            name = self.get_attrib('name')
        if not name:
            self.raise_error('"name" attribute is null')
        return name

    def get_name_and_value(self, ctx):
        name = self.get_name(ctx)
        if self.has_attrib('expr'):
            return name, self.eval_attrib(ctx, 'expr')
        value = self.get_attrib('value')
        if value is not None:
            return name, value
        else:
            return name, ctx.get_value(name)

# ------------------------------------------------------------------

# The HTML <INPUT> tag is wrapped so that variable values can be
# automatically pulled out of the execution context and rendered as
# the VALUE attribute.  When using Albatross application objects, the
# corresponding browser request is automatically merged back into the
# execution context.
class Input(EmptyTag, EvalAttribMixin):

    name = 'al-input'

    def __init__(self, ctx, filename, line_num, attribs):
        EmptyTag.__init__(self, ctx, filename, line_num, attribs)
        self.__itype = self.get_attrib('type', 'text').lower()
        try:
            self.unbound_to_html = self.type_dict[self.__itype]
        except KeyError:
            self.raise_error('unrecognised <INPUT TYPE="%s">' % self.__itype)
        self.assert_has_attribs('name', 'nameexpr', 'alias',
                                'prevpage', 'nextpage', 
                                'treefold', 'treeselect', 'treeellipsis')

    def write_attribs_except(self, ctx, *attribs):
        EmptyTag.write_attribs_except(self, ctx,
               'name', 'nameexpr', 'alias',
               'expr', 'valueexpr',
               'list', 'noescape',
               'nextpage', 'prevpage',
               'treefold', 'treeselect', 'treeellipsis', 'node',  *attribs)

    def get_name(self, ctx):
        for name in ('prevpage', 'nextpage'):
            if self.has_attrib(name):
                iter_name = self.get_attrib(name)
                return '%s,%s' % (name, iter_name)
        for name in ('treefold', 'treeselect', 'treeellipsis'):
            if self.has_attrib(name):
                iter_name = self.get_attrib(name)
                if self.has_attrib('node'):
                    node = self.eval_attrib(ctx, 'node')
                else:
                    iter = ctx.get_value(iter_name)
                    node = iter.value()
                return '%s,%s,%s' % (name, iter_name, node.albatross_alias())
        return EvalAttribMixin.get_name(self, ctx)

    def get_name_value_checked(self, ctx, default = None):
        # Name
        name = self.get_name(ctx)
        # Value
        if self.has_attrib('expr'):
            value = self.eval_attrib(ctx, 'expr')
        else:
            value = ctx.get_value(name)
        # Checked value
        if self.has_attrib('valueexpr'):
            return name, value, self.eval_attrib(ctx, 'valueexpr')
        elif self.has_attrib('value'):
            return name, value, self.get_attrib('value')
        else:
            return name, value, default


    def to_html(self, ctx):
        ctx.write_content('<input')
        self.unbound_to_html(self, ctx)
        ctx.write_content(' />')

    def generic_to_html(self, ctx):
        self.write_attribs_except(ctx, 'value')
        name, value = self.get_name_and_value(ctx)
        if name:
            ctx.write_content(' name="%s"' % name)
        if value is not None:
            if self.has_attrib('noescape'):
                ctx.write_content(' value="%s"' % value)
            else:
                ctx.write_content(' value="%s"' % escape(value))
        if not self.has_attrib('disabled'):
            ctx.input_add(self.__itype, name, value, self.has_attrib('list'))

    def radio_to_html(self, ctx):
        self.write_attribs_except(ctx, 'value', 'checked')
        name, value, value_attr = self.get_name_value_checked(ctx)
        ctx.write_content(' name="%s"' % name)
        ctx.write_content(' value="%s"' % value_attr)
        if str(value) == str(value_attr):
            ctx.write_content(' checked')
        if not self.has_attrib('disabled'):
            ctx.input_add('radio', name, value_attr, self.has_attrib('list'))

    def checkbox_to_html(self, ctx):
        self.write_attribs_except(ctx, 'value', 'checked')
        name, value, value_attr = self.get_name_value_checked(ctx, 'on')
        if value and self.has_attrib('treeselect'):
            value = 'on'
        ctx.write_content(' name="%s"' % name)
        ctx.write_content(' value="%s"' % value_attr)
        if type(value) in (type([]), type(())):
            if str(value_attr) in map(str, value):
                ctx.write_content(' checked')
        elif str(value) and str(value) == str(value_attr):
            ctx.write_content(' checked')
        if not self.has_attrib('disabled'):
            ctx.input_add('checkbox', name, value_attr, self.has_attrib('list'))

    def generic_novalue_to_html(self, ctx):
        # value= not applicable
        self.write_attribs_except(ctx, 'value')
        name = self.get_name(ctx)
        ctx.write_content(' name="%s"' % name)
        if not self.has_attrib('disabled'):
            ctx.input_add(self.__itype, name, None, self.has_attrib('list'))

    type_dict = {
        'text': generic_to_html, 'password': generic_to_html,
        'radio': radio_to_html, 'checkbox': checkbox_to_html,
        'submit': generic_to_html, 'reset': generic_to_html,
        'image': generic_novalue_to_html, 'file': generic_novalue_to_html,
        'hidden': generic_to_html, 'button': generic_to_html, }


# Allows the application need to modify the HREF attribute of the <A>
# tag to include extra information from the execution context.
class Href(EnclosingTag, EvalAttribMixin):

    name = 'al-a'

    def __init__(self, ctx, filename, line_num, attribs):
        EnclosingTag.__init__(self, ctx, filename, line_num, attribs)

    def to_html(self, ctx):
        ctx.write_content('<a')
        self.write_attribs_except(ctx, 'expr', 'href',
                                  'nextpage', 'prevpage',
                                  'treefold', 'treeselect', 'treeellipsis',
                                  'node')
        href = None
        for name in ('prevpage', 'nextpage'):
            if self.has_attrib(name):
                href = '%s?%s,%s=1' % (ctx.current_url(),
                                       name, self.get_attrib(name))
                break
        for name in ('treefold', 'treeselect', 'treeellipsis'):
            if self.has_attrib(name):
                iter_name = self.get_attrib(name)
                iter = ctx.get_value(iter_name)
                if self.has_attrib('node'):
                    node = self.eval_attrib(ctx, 'node')
                else:
                    node = iter.value()
                href = '%s?%s,%s,%s=1' % (ctx.current_url(),
                                          name, iter_name,
                                          node.albatross_alias())
                break
        if not href:
            if self.has_attrib('expr'):
                href = self.eval_attrib(ctx, 'expr')
            else:
                href = self.get_attrib('href')
            pos = href.find('?')
            if pos < 0:
                if href.find('=') >= 0:
                    # make into request on current page
                    href = '%s?%s' % (ctx.current_url(), href)
                else:
                    # make in request from full base_url path
                    href = ctx.redirect_url(href)
        ctx.write_content(' href="%s"' % href)
        ctx.write_content('>')
        EnclosingTag.to_html(self, ctx)
        ctx.write_content('</a>')


# Allows the application need to modify the SRC attribute of the <IMG>
# tag to include extra information from the execution context.
class Img(EmptyTag, EvalAttribMixin):

    name = 'al-img'

    def __init__(self, ctx, filename, line_num, attribs):
        EmptyTag.__init__(self, ctx, filename, line_num, attribs)
        self.assert_has_attribs('expr')

    def to_html(self, ctx):
        ctx.write_content('<img')
        self.write_attribs_except(ctx, 'expr', 'src', 'noescape')
        src = self.eval_attrib(ctx, 'expr')
        if self.has_attrib('noescape'):
            ctx.write_content(' src="%s"' % src)
        else:
            ctx.write_content(' src="%s"' % escape(src))
        ctx.write_content(' />')


# Using the albatross <FORM> tag allows Albatross to record the
# contents of each form.  This is used to register valid browser
# requests with the toolkit.
#
# We use a content_trap to capture the enclosed content so we know before
# emiting the <form> element whether or not to use a multipart/form-data
# encoding (required to make <input type="file"> work correctly).
class Form(EnclosingTag):

    name = 'al-form'

    def to_html(self, ctx):
        ctx.push_content_trap()
        ctx.form_open()
        EnclosingTag.to_html(self, ctx)
        use_multipart_enc = ctx.form_close()
        content = ctx.pop_content_trap()
        ctx.write_content('<form')
        self.write_attribs_except(ctx)
        if not self.has_attrib('action'):
            ctx.write_content(' action="%s"' % ctx.current_url())
        if use_multipart_enc and not self.has_attrib('enctype'):
            ctx.write_content(' enctype="multipart/form-data"')
        ctx.write_content('>')
        ctx.write_content(content)
        ctx.write_content('</form>')


# Applications which need to know the list of valid <al-select>
# options will need to use <al-option> tags in template files.  The
# alternative is to define the option list in Python code.
# 
# Two forms implemented:
# <al-option>12</al-option>
# <al-option value="12">Arbitrary text</al-option>
# 
# In the first form, we need to evaluate the content of the tag before
# we output the start tag so we can determine whether or not to set
# the selected attribute on the option, hence the use of the content_trap.
class Option(EnclosingTag):

    name = 'al-option'

    def __init__(self, ctx, filename, line_num, attribs):
        EnclosingTag.__init__(self, ctx, filename, line_num, attribs)

    def to_html(self, ctx, select_value):
        if self.has_attrib('value'):
            value = self.get_attrib('value')
        else:
            ctx.push_content_trap()
            EnclosingTag.to_html(self, ctx)
            value = ctx.pop_content_trap()
        ctx.write_content('<option')
        self.write_attribs_except(ctx)
        if type(select_value) is type(''):
            if value == select_value:
                ctx.write_content(' selected')
        elif type(select_value) in (type([]), type(())):
            if value in select_value:
                ctx.write_content(' selected')
        ctx.write_content('>')
        if self.has_attrib('value'):
            EnclosingTag.to_html(self, ctx)
        else:
            ctx.write_content(value)
        ctx.write_content('</option>')


# Some <al-select> tags have their option list supplied by an
# application value.  There are two forms:
# <al-select name="var">
#  <al-option>12</al-option>
# </al-select>
# <al-select name="var" optionexpr="range(12)"/>
class Select(EnclosingTag, EvalAttribMixin):

    name = 'al-select'

    def __init__(self, ctx, filename, line_num, attribs):
        EnclosingTag.__init__(self, ctx, filename, line_num, attribs)
        self.assert_has_attribs('name', 'alias', 'nameexpr')
        self.options = []

    def to_html(self, ctx):
        name, value = self.get_name_and_value(ctx)
        ctx.write_content('<select')
        self.write_attribs_except(ctx, 'optionexpr', 'alias', 'name',
                                  'nameexpr', 'noescape')
        ctx.write_content(' name="%s">' % name)
        if type(value) in (type([]), type(())):
            value = map(str, value)
        else:
            value = str(value)
        if self.has_attrib('optionexpr'):
            seq = self.eval_attrib(ctx, 'optionexpr')
            for item in seq:
                self.option_to_html(ctx, item, value)
        else:
            for item in self.options:
                item.to_html(ctx, value)
        ctx.write_content('</select>')
        if not self.has_attrib('disabled'):
            ctx.input_add('select', name, None, 
                          self.has_attrib('list') or self.has_attrib('multiple'))

    def option_to_html(self, ctx, item, select_value):
        if type(item) is type(()):
            value, text = map(str, item)
        else:
            value = str(item)
            text = value
        ctx.write_content('<option')
        if type(select_value) is type(''):
            if value == select_value:
                ctx.write_content(' selected')
        elif type(select_value) in (type([]), type(())):
            if value in select_value:
                ctx.write_content(' selected')
        if type(item) is type(()):
            ctx.write_content(' value="%s"' % escape(value))
        ctx.write_content('>')
        if self.has_attrib('noescape'):
            ctx.write_content(text)
        else:
            ctx.write_content(escape(text))
        ctx.write_content('</option>')
        ctx.write_content('\n')

    def append(self, tag):
        if isinstance(tag, Option):
            self.options.append(tag)
        else:
            EnclosingTag.append(self, tag)


class TextArea(EnclosingTag, EvalAttribMixin):

    name = 'al-textarea'

    def __init__(self, ctx, filename, line_num, attribs):
        EnclosingTag.__init__(self, ctx, filename, line_num, attribs)
        self.assert_has_attribs('name', 'alias', 'nameexpr')

    def to_html(self, ctx):
        name = self.get_name(ctx)
        ctx.write_content('<textarea')
        self.write_attribs_except(ctx, 'alias', 'name', 'nameexpr', 'list',
                                  'noescape')
        ctx.write_content(' name="%s">' % name)
        if ctx.has_value(name):
            value = ctx.get_value(name)
            if not value:
                ctx.write_content('')
            elif self.has_attrib('noescape'):
                ctx.write_content(value)
            else:
                ctx.write_content(escape(value))
        else:
            EnclosingTag.to_html(self, ctx)
        ctx.write_content('</textarea>')
        if not self.has_attrib('disabled'):
            ctx.input_add('textarea', name, None, self.has_attrib('list'))


# ------------------------------------------------------------------

# Send HTML accumulated in execution context to output.
class Flush(EmptyTag):

    name = 'al-flush'

    def to_html(self, ctx):
        ctx.flush_content()


# Do not execute enclosed content
class Comment(EnclosingTag):

    name = 'al-comment'

    def to_html(self, ctx):
        pass


# Include a template file. Can name the file directly, or can use an
# expression to yield a filename.
class Include(EmptyTag, EvalAttribMixin):

    name = 'al-include'

    def __init__(self, ctx, filename, line_num, attribs):
        EmptyTag.__init__(self, ctx, filename, line_num, attribs)
        self.assert_has_attribs('name', 'expr')

    def to_html(self, ctx):
        if self.has_attrib('expr'):
            name = self.eval_attrib(ctx, 'expr')
        else:
            name = self.get_attrib('name')
        templ = ctx.load_template(name)
        templ.to_html(ctx)


# Conditional content - emulates Python
#     if expr:
#     elif expr:
#     else:
#
# <al-if expr="value > 12">
#  Value > 12
# <al-elif expr="value > 5">
#  Value > 5 and <= 12
# <al-else>
#  Value <= 5
# </al-if>
class Elif(EmptyTag, EvalAttribMixin):

    name = 'al-elif'

    def __init__(self, ctx, filename, line_num, attribs):
        EmptyTag.__init__(self, ctx, filename, line_num, attribs)
        self.assert_has_attribs('expr')

    def is_true(self, ctx):
        return self.eval_attrib(ctx, 'expr')


class Else(EmptyTag):

    name = 'al-else'


class If(EnclosingTag, EvalAttribMixin):

    name = 'al-if'

    def __init__(self, ctx, filename, line_num, attribs):
        EnclosingTag.__init__(self, ctx, filename, line_num, attribs)
        self.assert_has_attribs('expr')
        self.elif_list = []
        self.else_content = not_specified
        self.current_content = self.content

    def to_html(self, ctx):
        if self.eval_attrib(ctx, 'expr'):
            EnclosingTag.to_html(self, ctx)
            return
        for tag, content in self.elif_list:
            if tag.is_true(ctx):
                content.to_html(ctx)
                return
        self.else_content.to_html(ctx)

    def append(self, tag):
        if isinstance(tag, Elif):
            self.current_content = Content()
            self.elif_list.append((tag, self.current_content))
        elif isinstance(tag, Else):
            self.current_content = Content()
            self.else_content = self.current_content
        else:
            self.current_content.append(tag)


# Execute some Python code
#
# <al-exec expr="keys = os.environ.keys(); keys.sort()">
class Exec(EmptyTag, EvalAttribMixin):

    name = 'al-exec'

    def __init__(self, ctx, filename, line_num, attribs):
        EmptyTag.__init__(self, ctx, filename, line_num, attribs)
        self.assert_has_attribs('expr')

    def to_html(self, ctx):
        self.eval_attrib(ctx, 'expr', 'exec')


# Iterated content - emulates Python
#     for iter in sequence:
class ListIterator:

    def __init__(self):
        self._index = self._start = self._count = 0
        self._seq = self._have_value = None
        self._pagesize = 0

    def __getstate__(self):
        return self._start, self._pagesize

    def __setstate__(self, tup):
        self._start, self._pagesize = tup
        self._index = self._start
        self._have_value = self._seq = None

    def __len__(self):
        if self._pagesize:
            return self._pagesize
        elif self._seq is not None:
            return len(self._seq)
        else:
            return 0

    # ----------------------------------------------------
    # Pagination functionality.

    def pagesize(self):
        return self._pagesize

    def set_pagesize(self, size):
        self._pagesize = size

    def has_prevpage(self):
        return self._start > 0

    def has_nextpage(self):
        if not self._pagesize:
            raise ApplicationError('iterator not in page mode')
        return len(self._seq) > self._start + self._pagesize

    def get_backdoor(self, op):
        return 1

    def set_backdoor(self, op, value):
        if value:
            if op == 'nextpage':
                self._start = self._start + self._pagesize
                self._index = self._start
            elif op == 'prevpage':
                self._start = self._start - self._pagesize
                if self._start < 0:
                    self._start = 0
                self._index = self._start

    # ----------------------------------------------------
    # Sequence control.

    def has_sequence(self):
        return self._seq is not None

    def set_sequence(self, seq):
        self._seq = seq

    # ----------------------------------------------------
    # Iteration.  Make sure that each element is only retrieved from
    # the sequence once.  This allows programs to make use of objects
    # which are not really sequences; popen().readline() for instance.

    def reset_index(self):
        self._index = self._start

    def reset_count(self):
        self._count = 0

    def index(self):
        return self._index

    def count(self):
        return self._count

    def start(self):
        return self._start

    def value(self):
        return self._value

    def clear_value(self):
        self._have_value = None

    def has_value(self):
        if self._have_value is None:
            try:
                self._value = self._seq[self._index]
                self._have_value = 1
            except IndexError:
                self._have_value = 0
        return self._have_value

    def set_value(self, value):
        self._value = value
        self._have_value = 1

    def next(self):
        self._index = self._index + 1
        self._count = self._count + 1
        self._have_value = None
        self.has_value()


# EXPR:seq
#   Yields an object which implements sequence protocol.  The result
#   is assigned to the ITER attribute.  If not present, the ITER
#   attribute must already have a sequence in it.
# PAGESIZE:int
#   Turns on pagination.  If this attribute is used anywhere with the
#   named ITER then that page size becomes the default for all other
#   uses of the ITER.  Places the ITER into the session.
# CONTINUE:bool
#   Instructs the tag not to reset the index() to the start() - it is
#   continued on from the previous ITER use.
# PREPARE:bool
#   Tells the tag to attach the sequence to the ITER and perform
#   pagination calculations.  This makes the ITER available for tests
#   before the content is rendered.
# COLS:int
#   Instructs the tag to render the content in multiple columns.
#   Items are arranged to flow down columns by default.
# FLOW:across/down
#   Define the order in which multicolumn output is arranged.
class For(EnclosingTag, EvalAttribMixin):

    name = 'al-for'

    def __init__(self, ctx, filename, line_num, attribs):
        EnclosingTag.__init__(self, ctx, filename, line_num, attribs)
        self.assert_has_attribs('iter')
        if self.has_attrib('flow'):
            if self.get_attrib('flow').lower() not in ('across', 'down'):
                self.raise_error('"flow" must be "across" or "down"')

    def to_html(self, ctx):
        # If PAGESIZE or CONTINUE defined then the iterator will not
        # be recreated if it already exists.
        name = self.get_attrib('iter')
        iter = None
        if self.has_attrib('pagesize') or self.has_attrib('continue'):
            iter = ctx.get_value(name)
        if iter is None:
            iter = ListIterator()
            ctx.set_value(name, iter)
        # If iterator does not have a sequence yet, the EXPR attribute
        # must be present
        if self.has_attrib('expr'):
            seq = self.eval_attrib(ctx, 'expr')
            iter.set_sequence(seq)
        elif not iter.has_sequence():
            self.raise_error('no sequence present; expected "expr" attribute')
        # If PAGESIZE is defined then place the iterator into the
        # session.
        if self.has_attrib('pagesize'):
            ctx.add_session_vars(name)
            pagesize = int(self.get_attrib('pagesize'))
            iter.set_pagesize(pagesize)
        # If CONTINUE if defined then the contents will flow on from
        # the previous use of the iterator
        if not self.has_attrib('continue'):
            iter.reset_index()
        iter.reset_count()
        iter.has_value()
        # If PREPARE defined then no rendering performed
        if self.has_attrib('prepare'):
            return
        # Render the sequence contents
        if self.has_attrib('cols'):
            if self.get_attrib('flow', 'down').lower() == 'down':
                self.cols_down_html(ctx, iter)
            else:
                self.cols_across_html(ctx, iter)
        else:
            pagesize = iter.pagesize()
            while iter.has_value():
                if pagesize and iter.count() >= pagesize:
                    break
                EnclosingTag.to_html(self, ctx)
                iter.next()
            iter.clear_value()

    def cols_down_html(self, ctx, iter):
        num_cols = int(self.get_attrib('cols'))
        num = len(iter)
        num_rows = num / num_cols
        num_extra = num % num_cols
        for row in range(num_rows + (num_extra > 0)):
            index = row
            seq = []
            for col in range(num_cols):
                if row == num_rows and col == num_extra:
                    break
                seq.append(iter._seq[index])
                index = index + num_rows + (col < num_extra)
                if index >= num:
                    break
            iter.set_value(seq)
            EnclosingTag.to_html(self, ctx)

    def cols_across_html(self, ctx, iter):
        num_cols = int(self.get_attrib('cols'))
        pagesize = iter.pagesize()
        num = len(iter)
        index = 0
        while index < num:
            seq = iter._seq[index:index + num_cols]
            iter.set_value(seq)
            EnclosingTag.to_html(self, ctx)
            index = index + num_cols


class TreeIterator:

    def __init__(self):
        self._stack = []

    def __setstate__(self, *tup):
        self._stack = []

    def tree_depth(self):
        return self._tree_depth

    def depth(self):
        return len(self._stack)

    def span(self):
        return self._tree_depth - len(self._stack)

    def line(self, depth):
        return self._line[depth]

    def set_line(self, line):
        self._line = line

    def value(self):
        return self._value

    def set_value(self, node):
        self._value = node

    def is_open(self):
        return hasattr(self._value, 'children')

    def node_use_ellipsis(self, ctx, node):
        return 0

    def use_ellipsis(self):
        return 0

    def node_type(self):
        return 0

    def node_is_open(self, ctx, node):
        return hasattr(node, 'children')

    def is_selected(self):
        return 0

    def has_children(self):
        return hasattr(self._value, 'children')

    def get_backdoor(self, op, key):
        return None

    def set_backdoor(self, op, key, value):
        pass


class LazyTreeIterator(TreeIterator):

    def __init__(self, single_select = 0):
        TreeIterator.__init__(self)
        self._single_select = single_select
        self._key = None
        self._type = 0
        self._selected_aliases = {}
        self._open_aliases = {}

    def __getstate__(self):
        return self._single_select, self._open_aliases, self._selected_aliases

    def __setstate__(self, tup):
        TreeIterator.__setstate__(self, tup)
        self._key = None
        self._single_select, self._open_aliases, self._selected_aliases = tup

    def set_value(self, node, type = 0):
        TreeIterator.set_value(self, node)
        if node:
            self._key = node.albatross_alias()
        self._type = type

    def is_open(self):
        if not hasattr(self._value, 'children'):
            return 0
        return self._open_aliases.has_key(self._key)

    def load_children(self, ctx, node):
        if not node.children_loaded:
            node.load_children(ctx)
            node.children_loaded = 1

    def node_is_open(self, ctx, node):
        if not hasattr(node, 'children'):
            return 0
        node_open = self._open_aliases.has_key(node.albatross_alias())
        if node_open:
            self.load_children(ctx, node)
        return node_open

    def is_selected(self):
        return self._selected_aliases.has_key(self._key)

    def node_type(self):
        return self._type

    def close_all(self):
        self._open_aliases = {}

    def deselect_all(self):
        self._selected_aliases = {}

    def get_selected_aliases(self):
        aliases = self._selected_aliases.keys()
        aliases.sort()
        return aliases

    def select_alias(self, alias, value = 1):
        if value:
            self._selected_aliases[alias] = 1
        elif self._selected_aliases.has_key(alias):
            del self._selected_aliases[alias]
        
    def set_selected_aliases(self, aliases):
        self._selected_aliases = {}
        for alias in aliases:
            self._selected_aliases[alias] = 1

    def get_open_aliases(self):
        aliases = self._open_aliases.keys()
        aliases.sort()
        return aliases

    def open_alias(self, alias, open = 1):
        if open:
            self._open_aliases[alias] = 1
        elif self._open_aliases.has_key(alias):
            del self._open_aliases[alias]

    def set_open_aliases(self, aliases):
        self._open_aliases = {}
        for alias in aliases:
            self._open_aliases[alias] = 1

    def get_backdoor(self, op, key):
        if op == 'treeselect':
            return self._selected_aliases.has_key(key)
        elif op == 'treefold':
            return 1
        else:
            return None

    def set_backdoor(self, op, key, value):
        if op == 'treeselect':
            if value:
                if self._single_select:
                    self._selected_aliases = { key: 1 }
                else:
                    self._selected_aliases[key] = 1
            elif self._selected_aliases.has_key(key):
                del self._selected_aliases[key]
        elif op == 'treefold':
            if value:
                if self._open_aliases.has_key(key):
                    del self._open_aliases[key]
                else:
                    self._open_aliases[key] = 1


class EllipsisTreeIterator(LazyTreeIterator):

    def __init__(self):
        LazyTreeIterator.__init__(self)
        self._noellipsis_alias = None

    def __setstate__(self, tup):
        LazyTreeIterator.__setstate__(self, tup)
        self._noellipsis_alias = None

    def node_use_ellipsis(self, ctx, node):
        if self._key == self._noellipsis_alias:
            return 0
        else:
            subnodes_open = 0
            for sub_node in self.value().children:
                if self.node_is_open(ctx, sub_node):
                    subnodes_open = 1
            return subnodes_open

    def get_backdoor(self, op, key):
        if op == 'treeellipsis':
            return 1
        else:
            return LazyTreeIterator.get_backdoor(self, op, key)

    def set_backdoor(self, op, key, value):
        if op == 'treeellipsis':
            if value:
                self._noellipsis_alias = key
        else:
            LazyTreeIterator.set_backdoor(self, op, key, value)


# Format tree structured data.
class Tree(EnclosingTag, EvalAttribMixin):

    name = 'al-tree'

    def __init__(self, ctx, filename, line_num, attribs):
        EnclosingTag.__init__(self, ctx, filename, line_num, attribs)
        self.assert_has_attribs('expr')
        self.assert_has_attribs('iter')

    def to_html(self, ctx):
        # Get iterator out of context - allows tree state to be
        # restored from session
        name = self.get_attrib('iter')
        iter = ctx.get_value(name)
        if iter is None:
            if self.has_attrib('ellipsis'):
                iter = EllipsisTreeIterator()
            elif self.has_attrib('lazy'):
                iter = LazyTreeIterator(self.has_attrib('single'))
            else:
                iter = TreeIterator()
            ctx.set_value(name, iter)
            ctx.add_session_vars(name)
        tree = self.eval_attrib(ctx, 'expr')
        iter._tree_depth = self.find_tree_depth(ctx, iter, tree)
        self.node_to_html(ctx, iter, tree)

    def find_tree_depth(self, ctx, iter, tree):
        deepest = 0
        iter.set_value(tree)
        if iter.node_is_open(ctx, tree):
            for node in tree.children:
                depth = self.find_tree_depth(ctx, iter, node)
                if depth > deepest:
                    deepest = depth
        return deepest + 1

    def node_to_html(self, ctx, iter, tree):
        parent = None
        line = []
        for node in iter._stack:
            if parent:
                if node is parent.children[-1]:
                    line.append(0)
                else:
                    line.append(1)
            parent = node
        if parent:
            if tree is parent.children[-1]:
                line.append(2)
            else:
                line.append(1)
        iter.set_line(line)
        iter.set_value(tree)
        EnclosingTag.to_html(self, ctx)
        iter._stack.append(tree)
        if iter.node_is_open(ctx, tree):
            if iter.node_use_ellipsis(ctx, tree):
                skipped = 0
                for sub_node in tree.children:
                    if iter.node_is_open(ctx, sub_node):
                        if skipped:
                            iter.set_value(tree, 1)
                            EnclosingTag.to_html(self, ctx)
                            skipped = 0
                        self.node_to_html(ctx, iter, sub_node)
                    else:
                        skipped = 1
                if skipped:
                    iter.set_value(tree, 1)
                    EnclosingTag.to_html(self, ctx)
            else:
                for node in tree.children:
                    self.node_to_html(ctx, iter, node)
        del iter._stack[-1]


# Include execution context values
#
# <al-value expr="name">
# <al-value expr="name" date="%d %b %Y">
# <al-value expr="name" lookup="name_lookup">
#
# When present the DATE attribute specifies a format string which is
# passed to time.strftime() to format the expression value.  The
# LOOKUP value uses the expression value as an index into the named
# lookup table and renders the value as the corresponding table entry.
class Value(EmptyTag, EvalAttribMixin):

    name = 'al-value'

    def __init__(self, ctx, filename, line_num, attribs):
        EmptyTag.__init__(self, ctx, filename, line_num, attribs)
        self.assert_has_attribs('expr')

    def to_html(self, ctx):
        value = self.eval_attrib(ctx, 'expr')
        format = self.get_attrib('date')
        if format:
            value = time.strftime(format, time.localtime(value))
            ctx.write_content(value)
            return
        lookup_name = self.get_attrib('lookup')
        if lookup_name:
            lookup = ctx.get_lookup(lookup_name)
            if not lookup:
                self.raise_error('undefined lookup "%s"' % lookup_name)
            lookup.lookup_html(ctx, value)
            return
        if self.has_attrib('noescape'):
            ctx.write_content(str(value))
        else:
            ctx.write_content(escape(value))


# Specify one entry in a lookup table
#
# <al-item expr="">
#  The enclosed content is evaluated when the table is
#  accessed, not when the table is contructed.  The EXPR
#  attribute is evaluated when the table is evaluated.
#  This allows the application to establish a context in
#  which the item EXPR attributes can be resolved to a
#  constant.
# </al-item>
class Item(EnclosingTag, EvalAttribMixin):

    name = 'al-item'

    def __init__(self, ctx, filename, line_num, attribs):
        EnclosingTag.__init__(self, ctx, filename, line_num, attribs)
        self.assert_has_attribs('expr')


# Specify a lookup table for translating a value to arbitrary content
# 
# <al-lookup name="name">
#  <al-item expr="1">Content</al-item>
#  <al-item expr="'hello'">Content</al-item>
#  <al-item expr="value + 3">Content</al-item>
# </al-lookup>
class Lookup(EnclosingTag):

    name = 'al-lookup'

    def __init__(self, ctx, filename, line_num, attribs):
        EnclosingTag.__init__(self, ctx, filename, line_num, attribs)
        self.assert_has_attrib('name')
        ctx.register_lookup(self.get_attrib('name'), self)
        self.item_list = []

    def append(self, tag):
        if isinstance(tag, Item):
            self.item_list.append(tag)
        else:
            EnclosingTag.append(self, tag)

    def _build_item_dict(self, ctx):
        if not hasattr(self, 'item_dict'):
            dict = {}
            for item in self.item_list:
                dict[item.eval_attrib(ctx, 'expr')] = item
            self.item_dict = dict

    def to_html(self, ctx):
        self._build_item_dict(ctx)

    def lookup_html(self, ctx, value):
        self._build_item_dict(ctx)
        item = self.item_dict.get(value, self.content)
        item.to_html(ctx)


# ------------------------------------------------------------------
# Macro expansions
# 
# Usearg and Macro do not produce output when evaluated - only when
# expanded via <al-expand>
class Usearg(EmptyTag):

    name = 'al-usearg'

    def to_html(self, ctx):
        content = ctx.get_macro_arg(self.get_attrib('name'))
        # Peel one level of arguments off stack to evaluate enclosed
        # content then restore the stack afterwards.  Nested macros
        # cause infinite recursion otherwise.
        args = ctx.pop_macro_args()
        content.to_html(ctx)
        ctx.push_macro_args(args)


class Macro(EnclosingTag):

    name = 'al-macro'

    def __init__(self, ctx, filename, line_num, attribs):
        EnclosingTag.__init__(self, ctx, filename, line_num, attribs)
        self.assert_has_attrib('name')
        ctx.register_macro(self.get_attrib('name'), self)

    def to_html(self, ctx):
        pass

    def expand_html(self, ctx):
        EnclosingTag.to_html(self, ctx)


class Setarg(EnclosingTag):

    name = 'al-setarg'


class Expand(EnclosingTag):

    name = 'al-expand'

    def __init__(self, ctx, filename, line_num, attribs):
        EnclosingTag.__init__(self, ctx, filename, line_num, attribs)
        self.arg_dict = {}

    def append(self, tag):
        if isinstance(tag, Setarg):
            self.arg_dict[tag.get_attrib('name')] = tag
        else:
            EnclosingTag.append(self, tag)

    def to_html(self, ctx):
        macro = ctx.get_macro(self.get_attrib('name'))
        if not macro:
            self.raise_error('undefined macro "%s"' % self.get_attrib('name'))
        args = {}
        args[None] = self.content
        for name, value in self.arg_dict.items():
            args[name] = value
        ctx.push_macro_args(args)
        macro.expand_html(ctx)
        ctx.pop_macro_args()


tags = (Input, Href, Img, Form, Option, Select, TextArea,
        Flush, Comment, Include, Elif, Else, If, Exec, For, Tree,
        Value, Item, Lookup, Usearg, Macro, Setarg, Expand)



#
# Copyright 2002 by Object Craft P/L, Melbourne, Australia.
#
# LICENCE - see LICENCE file distributed with this software for details.
#
#
# Test CGI Request functionality
#
# $Id: request.py,v 1.3 2003/01/24 11:34:31 andrewm Exp $

import sys
import os
import unittest
import cgi
import re
import StringIO
import albatross
import albatross.cgiapp

class Page:
    def page_process(self, ctx):
        self.locals = ctx.locals

class App(albatross.SimpleApp):
    def __init__(self):
        self.page = Page()
        page_name = self.page.__class__.__name__
        albatross.SimpleApp.__init__(self,
                                     base_url = '.', 
                                     template_path = '.', 
                                     start_page = page_name,
                                     secret = '')
        self.register_page(page_name, self.page)

class SimpleRequestTest(unittest.TestCase):
    def test_request(self, request):
        app = App()
        saved_stdout, sys.stdout = sys.stdout, StringIO.StringIO()
        try:
            app.run(request)
        finally:
            sys.stdout = saved_stdout
        self.failUnless(app.page.locals.foo, 'bah')
        self.failUnless(app.page.locals.list, ['a','b'])

class CGIGetRequestTestCase(SimpleRequestTest):
    def runTest(self):
        environ = {
            'REQUEST_METHOD': 'GET',
            'QUERY_STRING': 'foo=bah&list=a&list=b',
        }
        fields = cgi.FieldStorage(environ=environ)
        self.test_request(albatross.cgiapp.Request(fields=fields))

class CGIPostRequestTestCase(SimpleRequestTest):
    def runTest(self):
        environ = {
            'REQUEST_METHOD': 'POST',
            'CONTENT_TYPE': 'application/x-www-form-urlencoded',
        }
        input_file = StringIO.StringIO("foo=bah&list=a&list=b")
        fields = cgi.FieldStorage(fp=input_file, environ=environ)
        self.test_request(albatross.cgiapp.Request(fields=fields))

class CGIFormRequestTestCase(SimpleRequestTest):
    input = """\
----------AXBYCZ
Content-Disposition: form-data; name="foo"

bah
----------AXBYCZ
Content-Disposition: form-data; name="list"

a
----------AXBYCZ
Content-Disposition: form-data; name="list"

b
----------AXBYCZ--
"""
    def runTest(self):
        environ = {
            'REQUEST_METHOD': 'POST',
            'CONTENT_TYPE': 'multipart/form-data; boundary=--------AXBYCZ',
        }
        input_file = StringIO.StringIO(self.input)
        fields = cgi.FieldStorage(fp=input_file, environ=environ)
        self.test_request(albatross.cgiapp.Request(fields=fields))

class FileRequestTest(unittest.TestCase):
    def test_request(self, input_str):
        boundary = '--------AXBYCZ'
        environ = {
            'REQUEST_METHOD': 'POST',
            'CONTENT_TYPE': 'multipart/form-data; boundary=--------AXBYCZ',
        }
        app = App()
        input_file = StringIO.StringIO(input_str % self.make_albform())
        fields = cgi.FieldStorage(fp=input_file, environ=environ)
        saved_stdout, sys.stdout = sys.stdout, StringIO.StringIO()
        try:
            app.run(albatross.cgiapp.Request(fields=fields))
        finally:
            sys.stdout = saved_stdout
        return app.page.locals.foo

    def make_albform(self):
        # trick albatross into generating an __albform__ value for us
        app = App()
        ctx = albatross.SimpleAppContext(app)
        ctx.push_content_trap()
        ctx.form_open()
        ctx.input_add('file','foo','',0)
        ctx.form_close()
        result = ctx.pop_content_trap()
        match = re.search('name="__albform__" value="([^"]+)"', result)
        if match:
            return match.group(1)
        return ""

class CGISingleFileRequestTestCase(FileRequestTest):
    input_str = """\
----------AXBYCZ
Content-Disposition: form-data; name="foo"; filename="baz"

bah
----------AXBYCZ
Content-Disposition: form-data; name="__albform__"

%s
----------AXBYCZ--
"""
    def runTest(self):
        foo = self.test_request(self.input_str)
        self.failUnless(len(foo), 1)
        self.failUnless(foo[0].file.read(), 'bah')
        self.failUnless(foo[0].filename, 'baz')


class CGIListFileRequestTestCase(FileRequestTest):
    input_str = """\
----------AXBYCZ
Content-Disposition: form-data; name="foo"; filename="baz"

bah
----------AXBYCZ
Content-Disposition: form-data; name="foo"; filename="blot"

bling
----------AXBYCZ
Content-Disposition: form-data; name="__albform__"

%s
----------AXBYCZ--
"""
    def runTest(self):
        foo = self.test_request(self.input_str)
        self.failUnless(len(foo), 2)
        self.failUnless(foo[0].file.read(), 'bah')
        self.failUnless(foo[0].filename, 'baz')
        self.failUnless(foo[1].file.read(), 'bling')
        self.failUnless(foo[1].filename, 'blot')

class CGIMultiFileRequestTestCase(FileRequestTest):
    input_str = """\
----------AXBYCZ
Content-Disposition: form-data; name="foo"
Content-Type: multipart/mixed; boundary=--------ZCYBXA

----------ZCYBXA
Content-Disposition: attachment; filename="baz"
Content-Type: text/plain

bah
----------ZCYBXA
Content-Disposition: attachment; filename="blot"

bling
----------ZCYBXA--
----------AXBYCZ
Content-Disposition: form-data; name="__albform__"

%s
----------AXBYCZ--
"""
    def runTest(self):
        foo = self.test_request(self.input_str)
        self.failUnless(len(foo), 2)
        self.failUnless(foo[0].file.read(), 'bah')
        self.failUnless(foo[0].filename, 'baz')
        self.failUnless(foo[1].file.read(), 'bling')
        self.failUnless(foo[1].filename, 'blot')


"""
Driver for the above tests
"""

class RequestSuite(unittest.TestSuite):
    def __init__(self):
        unittest.TestSuite.__init__(self)
        self.addTest(CGIGetRequestTestCase())
        self.addTest(CGIPostRequestTestCase())
        self.addTest(CGIFormRequestTestCase())
        self.addTest(CGISingleFileRequestTestCase())
        self.addTest(CGIListFileRequestTestCase())
        self.addTest(CGIMultiFileRequestTestCase())

def suite():
    return RequestSuite()

if __name__ == '__main__':
    unittest.main(defaultTest='suite')



import sys
sys.path.insert(0, '..')
import install
from install import Installer

i = Installer('alsamp/random', cgi = 1)
i.add_exec('randompage.py')
i.add_exec('pages/tree.py', 'pages/')
i.add_html('pages/tree.html', 'pages/')
i.add_exec('pages/paginate.py', 'pages/')
i.add_html('pages/paginate.html', 'pages/')
i.add_module('utils.py')
i.install()



#!/usr/bin/env python
# To use:
#       python setup.py install
#

import distutils, os, sys, re
from distutils.core import setup, Extension
from distutils.command.sdist import sdist

class PreReleaseCheck:
    def __init__(self, distribution):
        self.distribution = distribution
        self.check_rev('doc/albatross.tex', r'^\\release.*\{([^}]+)\}')
        self.check_rev('doc/installation.tex', 
                       r'^tar xz.*albatross-(.*)\.tar\.gz')
        self.check_rev('doc/installation.tex', r'{albatross-(.*)}')
        self.check_rev('albatross/__init__.py', r'__version__ = \'(.*)\'')

    def _extract_rev(self, filename, pattern):
        regexp = re.compile(pattern)
        match = None
        revs = []
        line_num = 0
        f = open(filename)
        try:
            for line in f.readlines():
                line_num += 1
                match = regexp.search(line)
                if match:
                    revs.append((line_num, match.group(1)))
        finally:
            f.close()
        return revs

    def check_rev(self, filename, pattern):
        file_revs = self._extract_rev(filename, pattern)
        if not file_revs:
            sys.exit("Could not locate version in %s" % filename)
        line_num, file_rev = file_revs[0]
        for num, rev in file_revs[1:]:
            if rev != file_rev:
                sys.exit("%s:%d inconsistent version on line %d" % \
                         (filename, line_num, num))
        setup_rev = self.distribution.get_version()
        if file_rev != setup_rev:
            sys.exit("%s:%d version %s does not match setup.py version %s" % \
                     (filename, line_num, file_rev, setup_rev))


class my_sdist(sdist):
    def run(self):
        PreReleaseCheck(self.distribution)
        self.announce("Pre-release checks pass!")
        sdist.run(self)


setup(name = "albatross",
    version = "1.11",
    maintainer = "Object Craft",
    maintainer_email = "albatross@object-craft.com.au",
    description = "Albatross Web Toolkit",
    url = "http://www.object-craft.com.au/projects/albatross/",
    packages = ['albatross'],
    scripts = ['session-server/al-session-daemon',
               'standalone-server/al-httpd'],
    license = 'BSD - see file LICENCE',
    cmdclass = {'sdist': my_sdist},
)




#
# Copyright 2001 by Object Craft P/L, Melbourne, Australia.
#
# LICENCE - see LICENCE file distributed with this software for details.
#
# 
# Include functionality tests.
#
# $Id: include.py,v 1.6 2003/01/24 11:34:31 andrewm Exp $

import unittest
import albatross
import albatross_test

class IncludeCase(albatross_test.AlbatrossTestCase):
    template_dir = "include"

    def check_name(self):
        self.html_test("name.html", "plain\n")

    def check_expr(self):
        self.ctx.locals.filename = "name.html"
        self.html_test("expr.html", "plain\n")

    def check_nested(self):
        self.html_test("nested.html", "plain\n")

    def check_nxfile(self):
        self.assertRaises(albatross.ApplicationError, self.html_test, "incnxfile.html", "(none)")

class IncludeSuite(unittest.TestSuite):
    test_list = (
        "check_name",
        "check_expr",
        "check_nested",
        "check_nxfile",
    )
    def __init__(self):
        unittest.TestSuite.__init__(self, map(IncludeCase, self.test_list))

def suite():
    return IncludeSuite()

if __name__ == '__main__':
    unittest.main(defaultTest='suite')



#
# Copyright 2001 by Object Craft P/L, Melbourne, Australia.
#
# LICENCE - see LICENCE file distributed with this software for details.
#

import socket
import base64
import os.path

try:
    import Cookie
except ImportError:
    pass
try:
    import zlib
    have_zlib = 1
except ImportError:
    have_zlib = 0

from albatross.context import SessionBase
from albatross.common import *


# Simple session server
class SessionServerContextMixin(SessionBase):

    def __init__(self):
        SessionBase.__init__(self)
        self.__sesid = None

    def _get_sesid_from_cookie(self):
        hdr = self.request.get_header('Cookie')
        if hdr:
            c = Cookie.SimpleCookie(hdr)
            try:
                return c[self.app.ses_appid()].value
            except KeyError:
                pass
        return None

    def sesid(self):
        return self.__sesid

    def load_session(self):
        sesid = self._get_sesid_from_cookie()
        text = None
        if sesid:
            text = self.app.get_session(sesid)
        if not text:
            sesid = self.app.new_session()
        self.__sesid = sesid
        if text:
            text = base64.decodestring(text)
            try:
                if have_zlib:
                    text = zlib.decompress(text)
                self.decode_session(text)
            except:
                self.app.del_session(sesid)
                raise
        c = Cookie.SimpleCookie()
        c[self.app.ses_appid()] = self.__sesid
        path = self.absolute_base_url()
        c[self.app.ses_appid()]['path'] = path
        prefix_len = len('Set-Cookie: ')
        self.set_header('Set-Cookie', str(c)[prefix_len:])

    def save_session(self):
        if self.should_save_session():
            text = self.encode_session()
            if have_zlib:
                text = zlib.compress(text)
            text = base64.encodestring(text)
            self.app.put_session(self.__sesid, text)

    def remove_session(self):
        SessionBase.remove_session(self)
        if self.__sesid is not None:
            self.app.del_session(self.__sesid)


class SessionServerAppMixin:

    def __init__(self, appid, server = 'localhost', port = 34343, age = 1800):
        self.__appid = appid
        self.__server = server
        self.__port = port
        self.__sock = None
        self.__age = age
        try:
            self._server_connect()
        except ServerError:
            pass

    def ses_appid(self):
        return self.__appid

    def _server_connect(self):
        if self.__sock:
            return
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.connect((self.__server, self.__port))
            self.__sock = sock
        except socket.error:
            raise ServerError('could not connect to session server')

    def _server_read(self):
        try:
            text = self.__sock.recv(16384)
            if text:
                return text
        except socket.error:
            pass
        self.__sock = None
        raise ServerError('session server disconnect')

    def _server_write(self, text):
        try:
            self.__sock.send(text)
            return
        except socket.error:
            pass
        self.__sock = None
        raise ServerError('session server disconnect')

    def _server_read_response(self):
        resp = self._server_read()
        if not resp.startswith('OK'):
            raise ServerError('Session server returned: %s' % resp)
        return resp[2:].strip()

    def get_session(self, sesid):
        self._server_connect()
        self._server_write('get %s %s\r\n' % (self.__appid, sesid))
        text = ''
        while 1:
            text = text + self._server_read()
            pos = text.find('\r\n')
            if pos >= 0:
                if not text.startswith('OK'):
                    return
                text = text[pos + 2:]
                break
        while 1:
            if text.endswith('\r\n\r\n'):
                return text
            text = text + self._server_read()

    def new_session(self):
        self._server_connect()
        self._server_write('new %s %s\r\n' % (self.__appid, self.__age))
        return self._server_read_response()

    def put_session(self, sesid, text):
        self._server_connect()
        self._server_write('put %s %s\r\n' % (self.__appid, sesid))
        self._server_read_response()
        self._server_write(text.replace('\n', '\r\n'))
        self._server_write('\r\n')
        self._server_read_response()

    def del_session(self, sesid):
        self._server_connect()
        self._server_write('del %s %s\r\n' % (self.__appid, sesid))
        self._server_read_response()



import sys
sys.path.insert(0, '..')
import install
from install import Installer

i = Installer('alsamp/sybase', apache = 1)
i.add_exec('sybase.py')
i.add_html('macros.html')
i.add_html('table-desc.html')
i.add_html('table-list.html')
i.add_html('table-view.html')
i.add_html('traceback.html')
i.add_data('htaccess', '.htaccess')
i.install()



#
# Copyright 2003 by Object Craft P/L, Melbourne, Australia.
#
# LICENCE - see LICENCE file distributed with this software for details.
#
# AUTHOR(S)
#       Matt Goodall <matt AT pollenation DOT net>

import fcgi
from albatross import cgiapp


class Request(cgiapp.Request):
                                                 
    def __init__(self, fields = None):
        self.__fcgi = fcgi.FCGI()
        if fields is None:
            fields = self.__fcgi.getFieldStorage()
        cgiapp.Request.__init__(self, fields)
                                                         
    def get_uri(self):
        return self.__fcgi.env.get('REQUEST_URI')
                                                                     
    def get_path_info(self):
        return self.__fcgi.env.get('PATH_INFO')

    def get_servername(self):
        return self.__fcgi.env.get('HTTP_HOST')

    def get_header(self, name):
        env_name = 'HTTP_' + name.upper().replace('-', '_')
        return self.__fcgi.env.get(env_name)
                                                     

def running():
    return fcgi.isFCGI()



#
# Copyright 2001 by Object Craft P/L, Melbourne, Australia.
#
# LICENCE - see LICENCE file distributed with this software for details.
#
# 
# Basic functionality tests.
#
# $Id: basic.py,v 1.15 2003/09/21 04:40:31 djc Exp $

import unittest
import albatross
import albatross_test
import time

class BasicCase(albatross_test.AlbatrossTestCase):
    template_dir = "basic"

    def check_file(self):
        self.html_test("file.html", "plain\n")

    def check_nxfile(self):
        self.assertRaises(albatross.ApplicationError, self.html_test, "nxfile.html", "(none)")

    def check_const_value(self):
        self.html_test("const-value.html", "27")

    def check_null_attr(self):
        self.html_test("null-attr.html", '<img alt="" ismap src="" />')

    def check_attr_quote(self):
        self.html_test("attr-quote.html", "abcdefghijkl")

    def check_value_no_expr(self):
        self.assertRaises(albatross.ApplicationError, self.html_test, "value-no-expr.html", "(none)")

    def check_value_no_expr_sp(self):
        self.assertRaises(albatross.ApplicationError, self.html_test, "value-no-expr-sp.html", "(none)")

    def check_var_value(self):
        self.ctx.locals.x = 17
        self.html_test("var-value.html", "17")

    def check_var_float(self):
        self.ctx.locals.x = 2.1
        self.html_test("var-value.html", "2.1")

    def check_var_string(self):
        self.ctx.locals.x = "abc"
        self.html_test("var-value.html", "abc")

    def check_var_escape(self):
        self.ctx.locals.x = '<img src="http://rude.pictures.r.us/">'
        self.html_test("var-value.html", "&lt;img src=&quot;http://rude.pictures.r.us/&quot;&gt;")

    def check_var_noescape(self):
        self.ctx.locals.x = '<img src="http://rude.pictures.r.us/">'
        self.html_test("var-value-noescape.html", '<img src="http://rude.pictures.r.us/">')

    def check_date(self):
        t = time.strftime("%A, %B %d, %Y, %I:%M%p", 
                          time.localtime(996113203.67031395))
        self.html_test("date.html", t)

    def check_comment(self):
        self.html_test("comment.html", "one\nfour\n")

    def check_exec(self):
        self.html_test("exec.html", " 2 3 5 7")

class BasicSuite(unittest.TestSuite):
    test_list = (
        "check_file",
        "check_nxfile",
        "check_const_value",
        "check_null_attr",
        "check_attr_quote",
        "check_value_no_expr",
        "check_value_no_expr_sp",
        "check_var_value",
        "check_var_float",
        "check_var_string",
        "check_var_escape",
        "check_var_noescape",
        "check_date",
        "check_comment",
        "check_exec",
    )
    def __init__(self):
        unittest.TestSuite.__init__(self, map(BasicCase, self.test_list))

def suite():
    return BasicSuite()

if __name__ == '__main__':
    unittest.main(defaultTest='suite')



#!/usr/bin/python
#
# Copyright 2001 by Object Craft P/L, Melbourne, Australia.
#
# LICENCE - see LICENCE file distributed with this software for details.
#
import socket
import select
import random
import time
import binascii
import struct
import bisect
import sys
import getopt
import errno
import traceback

# Max consecutive server code failures a client can trigger before we
# close our connection to them.
MAX_CLIENT_EXCEPTIONS = 3


class Session:

    def __init__(self, sesid, life):
        self.sesid = sesid
        self.life = life
        self.text = ''
        self.last_access = time.time()

    def die_time(self):
        return self.last_access + self.life


class SessionDict:

    def __init__(self):
        self.dict = {}
        self.usage = 0

    def new_session(self, sesid, life):
        ses = Session(sesid, life)
        self.dict[sesid] = ses
        return ses

    def set_session(self, sesid, text):
        if not self.dict.has_key(sesid):
            return
        ses = self.dict[sesid]
        self.usage = self.usage - len(ses.text) + len(text)
        ses.text = text
        ses.last_access = time.time()

    def get_session(self, sesid, reset_access = 1):
        ses = self.dict.get(sesid)
        if ses and reset_access:
            ses.last_access = time.time()
        return ses

    def del_session(self, sesid):
        if self.dict.has_key(sesid):
            self.usage = self.usage = len(self.dict[sesid].text)
            del self.dict[sesid]

    def has_session(self, sesid):
        return self.dict.has_key(sesid)


class SesLife:

    def __init__(self, sesid, die_time):
        self.sesid = sesid
        self.die_time = die_time

    def __cmp__(self, other):
        return cmp(self.die_time, other.die_time)


class Reaper:

    def __init__(self, sessions):
        self.sessions = sessions
        self.list = []

    def add(self, sesid, life):
        bisect.insort(self.list, SesLife(sesid, time.time() + life))

    def process(self):
        now = time.time()
        while self.list:
            event = self.list[0]
            if event.die_time > now:
                break
            del self.list[0]
            ses = self.sessions.get_session(event.sesid, 0)
            if ses:
                die_time = ses.die_time()
                if die_time <= now:
                    self.sessions.del_session(event.sesid)
                else:
                    event.die_time = die_time
                    bisect.insort(self.list, event)
        if self.list:
            event = self.list[0]
            return event.die_time - now


class Context:

    def __init__(self, log_file):
        self.log_file = log_file
        self.sessions = SessionDict()
        self.reaper = Reaper(self.sessions)

        self.read_files = []
        self.write_files = []

    def log(self, msg):
        if self.log_file:
            self.log_file.write(msg)
            self.log_file.flush()

    def add_write_file(self, file):
        self.write_files.append(file)

    def del_write_file(self, file):
        self.write_files.remove(file)

    def add_read_file(self, file):
        self.read_files.append(file)

    def del_read_file(self, file):
        self.read_files.remove(file)

    def del_file(self, file):
        try:
            self.read_files.remove(file)
        except ValueError:
            pass
        try:
            self.write_files.remove(file)
        except ValueError:
            pass

    def select(self):
        timeout = self.reaper.process()
        try:
            read_out, write_out, oob_out = \
                select.select(self.read_files, self.write_files, [], timeout)
        except select.error, (eno, estr):
            if eno in (errno.EAGAIN, errno.EINTR):
                return [], []
            raise
        return read_out, write_out

    def reaper_add(self, sesid, life):
        self.reaper.add(sesid, life)

    def new_session(self, sesid, life):
        return self.sessions.new_session(sesid, life)

    def del_session(self, sesid):
        return self.sessions.del_session(sesid)

    def get_session(self, sesid):
        ses = self.sessions.get_session(sesid)
        if ses:
            self.log("session get %s, %d bytes, %d total\n" % \
                (sesid, len(ses.text), self.sessions.usage))
        return ses

    def set_session(self, sesid, text):
        self.log("session put %s, %d bytes, %d total\n" % \
            (sesid, len(text), self.sessions.usage))
        return self.sessions.set_session(sesid, text)

    def has_session(self, sesid):
        return self.sessions.has_session(sesid)


COMMAND = 0
PUT = 1


class Client:

    def __init__(self, context, sock, addr):
        self.context = context

        self.sock = sock
        self.addr = addr
        self.sock.setblocking(0)
        self.input = ''
        self.output = []
        self.state = COMMAND
        self.error_count = 0
        self.context.log('new client: %s from %s\n' % (sock.fileno(), addr[0]))

    def __del__(self):
        self.context.log('del client: %s\n' % self.sock.fileno())

    def fileno(self):
        return self.sock.fileno()

    def do_read(self):
        try:
            str = self.sock.recv(16384)
        except socket.error:
            return 0
        if not str:
            return 0
        self.input = self.input + str
        while 1:
            if self.state == COMMAND:
                pos = self.input.find('\r\n')
                if pos < 0:
                    return 1
                command = self.input[:pos]
                self.input = self.input[pos + 2:]
                if not self.handle_command(command):
                    return 0
            elif self.state == PUT:
                pos = self.input.find('\r\n\r\n')
                if pos < 0:
                    return 1
                text = self.input[:pos]
                self.input = self.input[pos + 4:]
                self.context.set_session(self.key, text)
                self.write('OK\r\n')
                self.state = COMMAND

    def do_write(self):
        while self.output:
            str = self.output[0]
            try:
                num = self.sock.send(str)
            except socket.error:
                return 0
            if num == len(str):
                del self.output[0]
            else:
                self.output[0] = str[num:]
                return 1
        self.context.del_write_file(self)
        return 1

    def write(self, str):
        if not self.output:
            self.context.add_write_file(self)
        self.output.append(str)

    def make_sesid(self, app):
        while 1:
            try:
                text = open('/dev/urandom').read(8)
            except:
                text = struct.pack('d', random.random())
            sesid = binascii.hexlify(text)
            if not self.context.get_session((app, sesid)):
                return sesid

    def handle_command(self, str):
        self.context.log('client %s cmd: %s\n' % (self.sock.fileno(), str))
        try:
            words = str.split()
            if not words:
                return 1
            cmd = words[0].lower()
            if cmd == 'new' and len(words) == 3:
                # new app life\r\n -> sesid\r\n
                # generate a new session id
                app = words[1]
                try:
                    life = int(words[2])
                except:
                    self.write('ERROR bad life\r\n')
                    return 1
                sesid = self.make_sesid(words[1])
                self.write('OK %s\r\n' % sesid)
                key = (app, sesid)
                self.context.new_session(key, life)
                self.context.reaper_add(key, life)
            elif cmd == 'get' and len(words) == 3:
                # get app sesid\r\n -> text\r\n
                # get the data for a session
                app, sesid = words[1:]
                key = (app, sesid)
                ses = self.context.get_session(key)
                if not ses:
                    self.write('ERROR no such session\r\n')
                else:
                    self.write('OK - session follows\r\n')
                    self.write(ses.text)
                    self.write('\r\n\r\n')
            elif cmd == 'put' and len(words) == 3:
                # put app sesid\r\ntext\r\n -> OK\r\n
                # put session data
                app, sesid = words[1:]
                key = (app, sesid)
                if not self.context.has_session(key):
                    self.write('ERROR no such session\r\n')
                else:
                    self.key = key
                    self.state = PUT
                    self.write('OK - send data now, terminate with blank line\r\n')
            elif cmd == 'del' and len(words) == 3:
                # del app sesid\r\n -> OK\r\n
                # delete session
                app, sesid = words[1:]
                key = (app, sesid)
                ses = self.context.get_session(key)
                if not ses:
                    self.write('ERROR no such session\r\n')
                else:
                    self.context.del_session(key)
                    self.write('OK\r\n')
            elif cmd == 'quit' and len(words) == 1:
                return 0
            else:
                self.write('ERROR unrecognised command: %s\r\n' % str)
            if self.error_count > 0:
                self.error_count = self.error_count - 1
            return 1
        except:
            self.error_count = self.error_count + 1
            if self.error_count > MAX_CLIENT_EXCEPTIONS:
                return 0
            traceback.print_exc()
            self.write('ERROR session server failure - check logs\r\n')
            return 1


class Server:

    def __init__(self, port, log_file):
        self.context = Context(log_file)

        self.listen_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.listen_sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.listen_sock.setblocking(0)
        self.listen_sock.bind(('', port))
        self.listen_sock.listen(5)

        self.context.add_read_file(self.listen_sock)

    def select_loop(self):
        self.run = 1
        while self.run:
            read_out, write_out = self.context.select()
            for s in read_out:
                if s is self.listen_sock:
                    c, addr = self.listen_sock.accept()
                    self.context.add_read_file(Client(self.context, c, addr))
                    del c # Otherwise socket won't be GC'ed in a timely manner
                else:
                    if not s.do_read():
                        self.context.del_file(s)
            for s in write_out:
                if not s.do_write():
                    self.context.del_file(s)
            read_out = write_out = s = None

    def stop(self):
        # This allows signal handlers to ask us to stop
        self.run = 0


def usage():
    sys.stderr.write('usage: session-server.py [-h] [-p port] [-l log-file]\n')


if __name__ == '__main__':
    port = 34343
    log_file = None
    try:
        opts, args = getopt.getopt(sys.argv[1:],
                                   "hp:l:", ["help", "port=", "log="])
    except getopt.GetoptError:
        usage()
        sys.exit(2)
    for opt, arg in opts:
        if opt in ("-h", "--help"):
            usage()
            sys.exit()
        elif opt in ("-p", "--port"):
            port = int(arg)
        elif opt in ("-l", "--log"):
            log_file = open(arg, 'a')
    server = Server(port, log_file)
    server.select_loop()
